"""
Notolog Editor
An open-source Markdown editor built with Python.

File Details:
- Purpose: Markdown Syntax Highlighter class tailored for Notolog.

Story:
This module combines regular expression patterns with integrated code logic to provide balanced results
and reduce reliance on complex regular expressions. This approach is particularly useful for processing multiline
blocks, such as code blocks, where stable detection is crucial for accurate visual representation. Inaccurate
detection may lead to "blinking highlighting." While the author has attempted to cover most cases, the primary
goal was to meet the common needs expected from a markdown editor.

Repository: https://github.com/notolog/notolog-editor
Website: https://notolog.app
PyPI: https://pypi.org/project/notolog

Author: Vadim Bakhrenkov
Copyright: 2024-2025 Vadim Bakhrenkov
License: MIT License

For detailed instructions and project information, please see the repository's README.md.
"""

from PySide6.QtCore import Qt

from .main_highlighter import MainHighlighter
from . import TextBlockData

from typing import TYPE_CHECKING

import re

if TYPE_CHECKING:
    from typing import Union  # noqa: F401
    from PySide6.QtGui import QTextBlockUserData  # noqa: F401


class MdHighlighter(MainHighlighter):
    """
    Syntax highlighter class for the Markdown language
    """

    theme = {
        'rn': {},
        'a': {'color': 'darkBlue'},
        'b': {'color': 'blue', 'style': 'bold'},
        'boo': {'color': 'blue', 'style': 'bold'},
        'i': {'color': 'red', 'style': 'italic'},
        'iu': {'color': 'red', 'style': 'italic'},
        'bi': {'color': 'darkRed', 'style': {'bold', 'italic'}},
        'biu': {'color': 'darkRed', 'style': {'bold', 'italic'}},
        'h1': {'color': 'white', 'font_size_ratio': 2.4,
               'bg': {'color': 'darkCyan', 'pattern': Qt.BrushStyle.Dense2Pattern}},
        'h1_text': {'color': 'white', 'font_size_ratio': 2.4, 'style': 'bold',
                    'bg': {'color': 'darkCyan', 'pattern': Qt.BrushStyle.Dense4Pattern}},
        'h2': {'color': 'white', 'font_size_ratio': 2.1,
               'bg': {'color': 'darkCyan', 'pattern': Qt.BrushStyle.Dense3Pattern}},
        'h2_text': {'color': 'darkCyan', 'font_size_ratio': 2.1, 'style': 'bold',
                    'bg': {'color': 'darkCyan', 'pattern': Qt.BrushStyle.Dense6Pattern}},
        'h3': {'color': 'white', 'font_size_ratio': 1.8,
               'bg': {'color': 'darkCyan', 'pattern': Qt.BrushStyle.Dense4Pattern}},
        'h3_text': {'color': 'darkCyan', 'font_size_ratio': 1.8, 'style': 'bold'},
        'h4': {'color': 'white', 'font_size_ratio': 1.6,
               'bg': {'color': 'darkCyan', 'pattern': Qt.BrushStyle.Dense4Pattern}},
        'h4_text': {'color': 'darkCyan', 'font_size_ratio': 1.6, 'style': 'bold'},
        'h5': {'color': 'white', 'font_size_ratio': 1.4,
               'bg': {'color': 'darkCyan', 'pattern': Qt.BrushStyle.Dense4Pattern}},
        'h5_text': {'color': 'darkCyan', 'font_size_ratio': 1.4, 'style': 'bold'},
        'h6': {'color': 'white', 'font_size_ratio': 1.3,
               'bg': {'color': 'darkCyan', 'pattern': Qt.BrushStyle.Dense4Pattern}},
        'h6_text': {'color': 'darkCyan', 'font_size_ratio': 1.3, 'style': 'bold'},
        's': {'color': 'grey', 'style': 'strikethrough'},
        'u': {'color': 'black', 'style': 'underline'},
        'code': {'color': 'magenta', 'bg': {'color': 'magenta', 'pattern': Qt.BrushStyle.Dense6Pattern}},
        'codel': {'color': 'yellow', 'style': 'monospace',
                  'bg': {'color': 'darkMagenta', 'pattern': Qt.BrushStyle.Dense2Pattern}},
        'codelf': {'color': 'white', 'style': 'monospace',
                   'bg': {'color': 'magenta', 'pattern': Qt.BrushStyle.Dense2Pattern}},
        'code_lang': {'color': 'magenta', 'style': 'bold'},
        'code_indent': {'bg': {'color': 'pink', 'style': 'monospace', 'pattern': Qt.BrushStyle.Dense6Pattern}},
        'code_content': {'color': 'brown', 'style': 'monospace'},
        # Debug: 'bg': {'color': 'darkGrey', 'pattern': Qt.BrushStyle.Dense2Pattern}
        'wrong_indent': {'bg': {'color': 'red', 'pattern': Qt.BrushStyle.DiagCrossPattern}},
        'comment': {'color': 'grey', 'style': 'monospace',
                    'bg': {'color': 'lightGrey', 'pattern': Qt.BrushStyle.Dense6Pattern}},
        'table_h': {'style': ['bold', 'monospace'], 'bg': {'color': 'lightGrey'}},
        'table_d': {'style': 'monospace', 'bg': {'color': 'whiteSmoke'}},
        'img': {'color': 'green'},
        'ref': {'color': 'white', 'bg': {'color': 'green', 'pattern': Qt.BrushStyle.Dense3Pattern}},
        # 'ref_data': {'color': 'green', 'bg': {'color': 'yellow', 'pattern': Qt.BrushStyle.Dense6Pattern}},
        'abbr': {'color': 'white', 'bg': {'color': 'dodgerBlue', 'pattern': Qt.BrushStyle.SolidPattern}},
        'abbr_text': {'color': 'white', 'style': 'bold',
                      'bg': {'color': 'dodgerBlue', 'pattern': Qt.BrushStyle.SolidPattern}},
        'link': {'color': 'white', 'style': 'italic', 'bg': 'blue'},
        'list': {'color': 'white', 'bg': 'darkMagenta'},
        'list_text': {'color': 'darkMagenta'},
        'list_indent': {'bg': {'color': 'darkMagenta', 'pattern': Qt.BrushStyle.Dense6Pattern}},
        'hr': {'color': 'white', 'style': 'strikethrough', 'bg': {'color': 'darkOrange'}},
        'blockquote': {'color': 'white', 'bg': {'color': 'grey', 'pattern': Qt.BrushStyle.Dense2Pattern},
                       # Blockquote friendly elements inherited this background
                       'bg_inner': {'color': 'lightGrey', 'pattern': Qt.BrushStyle.Dense3Pattern}},
        'html': {'color': 'darkCyan', 'style': 'monospace'},
        'html_open': {'color': 'green', 'style': 'monospace'},
        'html_close': {'color': 'darkRed', 'style': 'monospace'},
        'html_comment': {'color': 'grey', 'style': 'monospace'},
        'emoji': {'color': 'white', 'bg': {'color': 'olive', 'pattern': Qt.BrushStyle.Dense3Pattern}},
        'todo': {'color': 'darkCyan', 'bg': {'color': 'yellow', 'pattern': Qt.BrushStyle.Dense5Pattern}},
        'coop1': {'color': 'red', 'style': 'monospace', 'bg': {'color': None}},
        'coop2': {'color': 'darkGreen', 'style': 'monospace', 'bg': {'color': None}},
        'coop3': {'color': 'darkOrange', 'style': 'monospace', 'bg': {'color': None}},
        'coop4': {'color': 'blue', 'style': 'monospace', 'bg': {'color': None}},
        'coop5': {'color': 'brown', 'style': 'monospace',
                  'bg': {'color': 'brown', 'pattern': Qt.BrushStyle.Dense7Pattern}},
    }
    # Keep it consistent with ini-file name, say 'md.ini', item prefix 'md_color_h1_text'
    theme_ini_prefix = 'md'

    """
    Elements order is matter.
    Say, i* first, b** second, bi*** third one-by-one to override prev token.
        i_, b__, bi___
        code `...`, code ```...```

    nth 0 is a whole string
    nth > 1 calculates as: start = end - length
    """
    re_rules = [
        # Empty line
        (r'^([\r\n]*?)$', 1, 'rn', 'rn', False, theme['rn'], None),
        # Blockquotes
        (r'^(\s*?)(\>+\s.*?)$', 2, 'blockquote', 'blockquote', False, theme['blockquote'], None),
        # List
        (r'^(\s+|\>{1}\s+|)([0-9]{1,}\.|[\*\+\-]{1})(?=\s)', 2, 'list', 'list', False, theme['list'], None),
        (r'^(\s+|\>{1}\s+|)([0-9]{1,}\.|[\*\+\-]{1})(\s.*?)$', 3, 'list_text', 'list', False, theme['list_text'], None),
        (r'^(\s+|\>{1}\s+|)([0-9]{1,}\.|[\*\+\-]{1})\s.*?$',
         1, 'list_indent', 'list', False, theme['list_indent'], None),
        # Code ``` (fenced)
        (r'(?:^|\s|\W|[^`\#])(?<!`)(```(?!`).*?(?<!`)```)(?!`)(?:\s|\W|[^`]|$)',
         1, 'codel', 'code', False, theme['codelf'], None),
        (r'^(?<!\#)((?:[\s]*?)```)[a-z\-_\+#\s\.{}]*?(?!```)$', 1, 'code', 'code', True, theme['code'], None),
        (r'^(?<!\#)((?:[\s]*?)```)(\s*?\{?[a-z\-_\+#\s\.^{}]+\}?)$',
         2, 'code_lang', 'code', False, theme['code_lang'], None),
        (r'^([\s]{1,})```([\S]+|)$', 1, 'wrong_indent', 'code', False, theme['wrong_indent'], None),
        # Code ::::
        (r'^(([\s]{4,}|[\t]{1,})(::::))[\S]*?\s*?$', 3, 'codec', 'code', False, theme['code'], None),
        (r'^(([\s]{4,}|[\t]{1,})::::)([\S]+)\s*?$', 3, 'code_lang', 'code', False, theme['code_lang'], None),
        (r'^([\s]{4,}|[\t]{1,})::::[\S]*?\s*?$', 1, 'code_indent', 'code', False, theme['code_indent'], None),
        # Italic (asterisk)
        (r'(?<!\*)(\*[^\s\*][^\*]*?[^\s]?\*)(?!\*)', 1, 'i', 'i', False, theme['i'], None),  # between
        # \W to avoid character mention in a sequence like: *, ...
        (r'(^|\s)(\*[^\s\W\*][^\*]*?)(?!\*)$', 2, 'i_open', 'i', True, theme['i'], None),
        (r'^([^\*]+(?<!\s)\*)(?!\*)(?:\s|\W|$)', 1, 'i_close', 'i', True, theme['i'], None),
        # Italic (underline)
        # \W (non-word) characters are all characters apart from numbers, letters, and underscores.
        (r'(^|\s)(_(?!\s|_).*?(?<!\s|_)_)(?:\s|[\W^_]|$)', 2, 'iu', 'iu', False, theme['iu'], None),  # between
        (r'(^|\s)(_[^\s\W_][^_]*?)$', 2, 'iu_open', 'iu', True, theme['iu'], None),
        (r'^([^_]*?(?<!_)[^\s_]_)(?!_)(?:\s|\W|$)', 1, 'iu_close', 'iu', True, theme['iu'], None),
        # Bold (asterisk)
        (r'(?<!\*)(\*{2}[^\s\*][^\*]*?[^\s]?\*{2})(?!\*)', 1, 'b', 'b', False, theme['b'], None),  # between
        (r'(^|\s)(\*{2}[^\s\*][^\*]*?)(?!\*)$', 2, 'b_open', 'b', True, theme['b'], None),
        (r'^([^\*]+(?<!\s)\*{2})(?!\*)(?:\s|\W|$)', 1, 'b_close', 'b', True, theme['b'], None),
        # Bold (underline)
        # \b doesn't work, only [^\s]
        (r'(^|\s)(__(?!\s|_).*?(?<!\s|_)__)(?:\s|[\W^_]|$)', 2, 'boo', 'boo', False, theme['boo'], None),  # between
        (r'(^|\s)(__[^\s\W_][^_]*?)$', 2, 'boo_open', 'boo', True, theme['boo'], None),
        (r'^([^_]*?(?<!_)[^\s_]__)(?!__)(?:\s|\W|$)', 1, 'boo_close', 'boo', True, theme['boo'], None),
        # (r'(?<=__)([^\s].*?[^\s])(?=__)', 'b', 'b', 1, True, theme['b'], None),
        # Bold and Italic altogether (asterisk)
        (r'(?<!\*)(\*{3}[^\s\*][^\*]*?[^\s]?\*{3})(?!\*)', 1, 'bi', 'bi', False, theme['bi'], None),  # between
        (r'(^|\s)(\*{3}[^\s\*][^\*]*?)(?!\*)$', 2, 'bi_open', 'bi', True, theme['bi'], None),
        (r'^([^\*]+(?<!\s)\*{3})(?!\*)(?:\s|\W|$)', 1, 'bi_close', 'bi', True, theme['bi'], None),
        # Bold and Italic altogether (underline)
        (r'(^|\s)(___(?!\s|_).*?(?<!\s|_)___)(?:\s|[\W^_]|$)', 2, 'biu', 'biu', False, theme['biu'], None),  # between
        (r'(^|\s)(___[^\s\W_][^_]*?)$', 2, 'biu_open', 'biu', True, theme['biu'], None),
        (r'^([^_]*?(?<!_)[^\s_]___)(?!___)(?:\s|\W|$)', 1, 'biu_close', 'biu', True, theme['biu'], None),
        # Strikethrough
        # ~~text~~ first to allow skip the open-close group as it may interference
        (r'(~~(?!~|\s)[^~]*?(?<!~|\s)~~)', 1, 's', 's', False, theme['s'], None),  # between
        (r'(^|\s)(?<!~~)(~~(?!~|\s)[^~]*?)$', 2, 's_open', 's', True, theme['s'], None),
        (r'^([^~]*?[^\s~]~~)(?!~~)(?:\s|[\W^~]|$)', 1, 's_close', 's', True, theme['s'], None),
        # Underline
        (r'(<u>.*?</u>)', 1, 'u', 'u', False, theme['u'], None),  # between
        (r'(<u>[^<>]*?)(?!</?u>)$', 1, 'u_open', 'u', True, theme['u'], None),
        (r'^([^<>]*?</u>)(?!</?u>)(?:\s|\W|$)', 1, 'u_close', 'u', True, theme['u'], None),
        # Code line within backticks
        (r'(?:^|\s|\W|[^`\#])(?<!`)(`(?!`).*?(?<!`)`)(?!`)(?:\s|\W|[^`]|$)',
         1, 'codel', 'code', False, theme['codel'], None),
        # Header
        # (r'^(?:[\s\t]*?)(?<h1>#\s*?)(?<h1_text>.*)$', ['h1', 'h1_text'],
        # 'h1_text', 'h1', False, theme['h_text'], None),
        (r'^((?:[\s\t]*?)#)\s*?.*?', 1, 'h1', 'h1', False, theme['h1'], None),
        (r'^((?:[\s\t]*?)#\s*?)(.*)$', 2, 'h1_text', 'h1', False, theme['h1_text'], None),
        (r'^((?:[\s\t]*?)[#]{2})\s*?.*?', 1, 'h2', 'h', False, theme['h2'], None),
        (r'^((?:[\s\t]*?)[#]{2}\s*?)(.*)$', 2, 'h2_text', 'h', False, theme['h2_text'], None),
        (r'^((?:[\s\t]*?)[#]{3})\s*?.*?', 1, 'h3', 'h', False, theme['h3'], None),
        (r'^((?:[\s\t]*?)[#]{3}\s*?)(.*)$', 2, 'h3_text', 'h', False, theme['h3_text'], None),
        (r'^((?:[\s\t]*?)[#]{4})\s*?.*?', 1, 'h4', 'h', False, theme['h4'], None),
        (r'^((?:[\s\t]*?)[#]{4}\s*?)(.*)$', 2, 'h4_text', 'h', False, theme['h4_text'], None),
        (r'^((?:[\s\t]*?)[#]{5})\s*?.*?', 1, 'h5', 'h', False, theme['h5'], None),
        (r'^((?:[\s\t]*?)[#]{5}\s*?)(.*)$', 2, 'h5_text', 'h', False, theme['h5_text'], None),
        (r'^((?:[\s\t]*?)[#]{6})\s*?.*?', 1, 'h6', 'h', False, theme['h6'], None),
        (r'^((?:[\s\t]*?)[#]{6}\s*?)(.*)$', 2, 'h6_text', 'h', False, theme['h6_text'], None),
        # Table
        (r'^(\|[\s\|\:\-]+?\|)$', 1, 'table_h', 'table', False, theme['table_h'], None),
        (r'^(\|(?=.*?[a-zA-Z0-9]).*?\|)$', 1, 'table_d', 'table', False, theme['table_d'], None),
        # Image
        (r'(!\[[^\]]*?\]\([^\)]*?\))', 1, 'img', 'img', False, theme['img'], None),
        (r'(!\[[^\]]*?\]\[[^\]]*?\])', 1, 'img', 'img', False, theme['img'], None),
        # reference either image or link, footnotes also
        (r'((?<!\*)\[[^\]]*?\]:)', 1, 'ref', 'ref', False, theme['ref'], None),
        # (r'(\[[^\]]*?\]:)(\S*?)$', 2, 'ref_data', 'ref', False, theme['ref_data'], None),
        # abbreviations
        (r'(\*\[.*?\]:.*?)', 1, 'abbr', 'abbr', False, theme['abbr'], None),
        (r'(\*\[(.*?)\]:.*?)', 2, 'abbr_text', 'abbr', False, theme['abbr_text'], None),
        # hyperlinks (before the web links block to allow style overriding)
        (r'((?<!!)\[.*?\]\(.*?\))', 1, 'a', 'a', False, theme['a'], None),
        # Web link
        # A word boundary \b matches the position between a word character (e.g., alphanumeric character or underscore)
        # and a non-word character (e.g., whitespace, punctuation, or the beginning/end of a string).
        # It allows you to match patterns only at the boundaries of words.
        # RFC 3986: uri = "[A-Za-z0-9\-._~:/?#\[\]@!$&\'()*+,;=%]"
        (r'((?:https?|ftp):\/\/(?:\S+(?::\S*)?@)?'
         r'(?:\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}|(?!-)[A-Za-z0-9-]{1,63}(?:\.(?!-)[A-Za-z0-9-]{1,63})+)?(?:\:[0-9]+)?'
         r'(?:\/(?:[^\:\?#\s\/\)]+)?)*(?:\?[^\s]*)?(?:#[^\s]*)?)', 1, 'link', 'link', False, theme['link'], None),
        # horizontal line
        (r'^(?:[\s>]*?)([\-\*_]{3,}(?:[\s]*?))$', 1, 'hr', 'hr', False, theme['hr'], None),
        # Comments
        # TODO: Add syntax highlighting based on the code language.
        (r'^([\s\t]*?[#]{1,}\s*?.*)$', 1, 'comment', 'comment', False, theme['comment'],
         lambda s: s.is_in_code()),
        (r'^((?:\s*?)"""(?!").*?(?<!")""")(?:\s*?)$', 1, 'comment', 'comment', False, theme['comment'],
         lambda s: s.is_in_code()),
        (r'^(?:[\s\t]*?)(""")(?:[\s\t]*?)$', 1, 'cclop', 'comment', False, theme['comment'],
         lambda s: s.is_in_code()),
        (r'(\/\/.*?)$', 1, 'comment', 'comment', False, theme['comment'],
         lambda s: s.is_in_code()),
        # html tags
        (r'(<\s*?([a-zA-Z][^>]*?)\s*?>)', 1, 'html', 'html', True, theme['html_open'], None),
        (r'(<\s*?\/\s*?([a-zA-Z][^\s>]*?)\s*?>)', 1, 'html', 'html', True, theme['html_close'], None),
        (r'(<\s*?([a-zA-Z!][^\/]*?)\s*?(\/)?\s*?>)', 1, 'html', 'html', False, theme['html'], None),
        (r'(<!--(.*?)-->)', 1, 'html', 'html', False, theme['html_comment'], None),
        # emojis
        (r'(\:[a-zA-Z_]+\:)', 1, 'emoji', 'emoji', False, theme['emoji'], None),
        # To-do keywords
        (r'([\s]*?)(@todo)(?=\s|$)', 2, 'todo', 'todo', False, theme['todo'], None),
        # Excess indent at the end of the line
        # (r'([\s]+)$', 1, 'wrong_indent', 'general', False, theme['wrong_indent'], None),
        # Code operators, group 1
        (r'(class|exit|[\-\+\*\{\}]+)', 1, 'coop1', 'coop', False, theme['coop1'],
         lambda s: s.is_in_code() and not s.is_in_code_comment()),
        # Assignments, group 2. Can be extended to [\=\[\]\(\)]
        # (r'([\=]{1,})', 1, 'coop2', 'coop', False, theme['coop2'],
        # lambda s: s.is_in_code() and not s.is_in_code_comment()),
        # Functions and methods, group 3
        (r'(\b(?:\s*?)[a-zA-Z\.\_]+\b)\(.*?\)', 1, 'coop3', 'coop', False, theme['coop3'],
         lambda s: s.is_in_code() and not s.is_in_code_comment()),
        # Code operators and instructions, group 4
        (r'((?:^|\s)*?if|else|elif|for|while|switch|def|function|lambda|echo|self|return)[\s\W]{1,}',
         1, 'coop4', 'coop', False, theme['coop4'],
         lambda s: s.is_in_code() and not s.is_in_code_comment()),
        # Numbers in code, group 5
        # (r'([0-9]+)', 1, 'coop5', 'coop', False, theme['coop5'],
        # lambda s: s.is_in_code() and not s.is_in_code_comment()),
    ]

    def get_regex(self, pattern: re) -> re:
        """
        Get either QRegularExpression or raw Python regex
        re = QRegularExpression(pattern)
        """
        return pattern

    def get_open_close_token_map(self):
        return [
            {'group': 'i', 'open': 'i_open', 'close': 'i_close', 'theme': 'i'},
            {'group': 'iu', 'open': 'iu_open', 'close': 'iu_close', 'theme': 'iu'},
            {'group': 'b', 'open': 'b_open', 'close': 'b_close', 'theme': 'b'},
            {'group': 'boo', 'open': 'boo_open', 'close': 'boo_close', 'theme': 'boo'},
            {'group': 'bi', 'open': 'bi_open', 'close': 'bi_close', 'theme': 'bi'},
            {'group': 'biu', 'open': 'biu_open', 'close': 'biu_close', 'theme': 'biu'},
            {'group': 's', 'open': 's_open', 'close': 's_close', 'theme': 's'},
            {'group': 'u', 'open': 'u_open', 'close': 'u_close', 'theme': 'u'}
        ]

    def is_blockquote_inner(self, tag=str):
        return tag in {
            'i', 'i_open', 'i_close', 'iu', 'iu_open', 'iu_close',
            'b', 'b_open', 'b_close', 'boo', 'boo_open', 'boo_close',
            'bi', 'bi_open', 'bi_close', 'biu', 'biu_open', 'biu_close',
            's', 's_open', 's_close',
            'u', 'u_open', 'u_close',
            'list_text', 'list_indent',
            'emoji',
            'hr',
            'a'
        }

    def get_nl_closing_tokens(self):
        """
        Block open tokens closing with a new empty line.
        The order in the row is matter because of processing one-by-one.
        """
        return ['rn', 'codec', 'blockquote', 'list']

    def highlightBlock(self, text_str):  # noqa: C901 - consider simplifying this method
        """
        Apply a syntax highlighting to each line of the text.
        * https://doc.qt.io/qt-6/qsyntaxhighlighter.html#highlightBlock
        """

        # Get the current block and associated user data
        self.current_block = self.currentBlock()
        self.user_data = self.current_block.userData()  # type: Union[TextBlockData, QTextBlockUserData]

        self.prev_block = self.current_block.previous()
        self.prev_user_data = self.prev_block.userData()  # type: Union[TextBlockData, QTextBlockUserData]

        self.line_number = self.currentBlock().blockNumber()
        # Line number as it appears in the editor
        self.line_number_log = self.line_number + 1

        if self.user_data is None or not isinstance(self.user_data, TextBlockData):
            self.logger.debug('{%r} !!! Block data is not set at [%d*], is in code %d, prev state %d'
                              % (self.rehighlight_block, self.line_number_log,
                                 self.is_in_code(skip_data=True), self.previousBlockState()))
            self.user_data = TextBlockData(self.line_number)
            """
            Restore within a code block state, for example when a new line appears.
            Both 'code' and 'codec' have their own rules on how a code group works, so check them up separately.
            """
            if self.is_in_code(skip_data=True, force_tag='code'):
                self.user_data.put(tag='code', opened=False, within=True, closed=True)
            elif self.is_in_code(skip_data=True, force_tag='codec'):
                self.user_data.put(tag='codec', opened=False, within=True, closed=True)

        # Keep all line numbers even if no tags there
        if self.line_number not in self.line_tokens:
            self.line_tokens[self.line_number] = {}

        # Each line is not formatted initially
        self.clear_formatted()

        # Groups of tokens for correction (they have similar approach in rules)
        oct_groups = [r.get('group') for r in self.get_open_close_token_map()]
        open_tokens = [r.get('open') for r in self.get_open_close_token_map()]
        close_tokens = [r.get('close') for r in self.get_open_close_token_map()]

        pattern, nth, tag, group, duple, cf_data, reckon = (None,) * 7
        for pattern, nth, tag, group, duple, cf_data, reckon in self.rules:
            """
            Process code block tokens.
            """
            if tag == 'code':
                if duple and tag not in self.tokens:
                    self.tokens[tag] = {'cnt': 0, 'o': None}
                # Pass params below
                break

        if re is None:
            self.logger.debug('Notice: Regex pattern not found for the tag "%s"' % tag)
            return

        if self.rehighlight_block:
            if 'code' in self.tokens:
                # Opened and continues to apply for non-code blocks
                self.tokens['code']['o'] = (self.user_data.get_param('code', 'within')
                                            and not self.user_data.get_param('code', 'closed'))

        matches = re.finditer(pattern, text_str)
        match = next(matches, None)

        if match:
            # Get regex result position into the text string
            start = match.start(nth)
            end = match.end(nth)
            length = end - start

            # Collect line tokens only when any of them matched
            if tag not in self.line_tokens[self.line_number]:
                self.line_tokens[self.line_number][tag] = []
            # The line tokens data will be reset after re-highlighting, no need to check for duplicates
            line_token_data = {'start': start, 'end': end, 'length': length}
            if line_token_data not in self.line_tokens[self.line_number][tag]:
                self.line_tokens[self.line_number][tag].append(line_token_data)

            if self.rehighlight_block:
                if tag in self.tokens:
                    """
                    Recover state of an open token
                    """
                    self.tokens[tag]['o'] = (self.user_data.get_param(tag, 'within')
                                             and self.user_data.get_param(tag, 'opened'))
            elif not self.rehighlight_block:
                # Set up the tag count
                if tag in self.tokens:
                    self.tokens[tag]['cnt'] += 1
                # Because of the code token could be the same whether opened or closed.
                self.tokens[tag]['o'] = True if self.tokens[tag]['cnt'] % 2 > 0 else False

            self.logger.debug(
                'Current code block data [%d] rehi:%r, curr_blk_st:%d, prev_blk_st:%d, '
                'in_code:%r, in_code:%r(STRICT), inc:%r, o:[%r]~[%r], c:%r'
                % (self.line_number,
                   self.rehighlight_block,
                   self.currentBlockState(),
                   self.previousBlockState(),
                   self.is_in_code(),
                   self.is_in_code(skip_data=True),
                   self.user_data.get_param(tag, 'within'),
                   self.user_data.get_param(tag, 'opened'),
                   self.tokens[tag]['o'],
                   self.user_data.get_param(tag, 'closed')))

            # Set the correct state of the code block to update currentBlockState()
            self.check_and_set_in_code_state()

            # Check either tag opened or closed and set a relevant data
            if self.tokens[tag]['o']:
                self.logger.debug(
                    '{%r} >>> Open "%s" at [%d*], is in code %d, prev state %d'
                    % (self.rehighlight_block, tag, self.line_number_log, self.is_in_code(),
                       self.previousBlockState())
                )
                if not self.user_data.get_param(tag, 'opened'):
                    self.user_data.put(tag=tag, opened=True, within=True, closed=False)
            # If tag wasn't open it doesn't mean it was closed, check prev state
            elif self.prev_user_data and self.prev_user_data.get_param(tag, 'within'):
                self.logger.debug(
                    '{%r} <<< Close "%s" at [%d*], is in code %d, prev state %d'
                    % (self.rehighlight_block, tag, self.line_number_log, self.is_in_code(),
                       self.previousBlockState())
                )
                # Close the code block
                if not self.user_data.get_param(tag, 'closed'):
                    self.user_data.put(tag=tag, opened=False, within=True, closed=True)
            else:
                # Warning as such case is not expected here
                self.logger.debug(
                    'Notice: {%r} >~> Open token inside "%s" at [%d*]'
                    % (self.rehighlight_block, tag, self.line_number_log)
                )
        elif self.is_in_code(skip_data=True, force_tag='code'):
            self.user_data.put(tag=tag, opened=False, within=True, closed=False)
            self.setCurrentBlockState(1)
            if not self.is_any_formatted():
                self.setFormat(0, len(text_str), self.cf(**self.theme['code_content']))
                self.set_formatted('code')
            self.logger.debug(
                '{%r} >=< Inside "%s" at [%d*], is in code %d (STRICT), prev state %d'
                % (self.rehighlight_block, tag, self.line_number_log,
                   self.is_in_code(skip_data=True, force_tag='code'), self.previousBlockState())
            )
        else:
            # When new line within a code block appears have to check prev block state (within and not closed)
            if (self.is_in_code(force_tag='code')
                    and self.prev_user_data
                    and self.prev_user_data.get_param(tag, 'within')
                    and not self.prev_user_data.get_param(tag, 'closed')):
                self.user_data.put(tag=tag, opened=False, within=True, closed=False)
                self.setCurrentBlockState(1)
                self.logger.debug(
                    '{%r} >=< Inside "%s" [%d*], is in code %d (lenient), prev state %d'
                    % (self.rehighlight_block, tag, self.line_number_log, self.is_in_code(force_tag='code'),
                       self.previousBlockState())
                )
            else:
                self.user_data.put(tag=tag, opened=False, within=False, closed=False)
                self.setCurrentBlockState(0)
                self.logger.debug('{%r} ... No "%s" [%d*]' % (self.rehighlight_block, tag, self.line_number_log))

        for pattern, nth, tag, group, duple, cf_data, reckon in self.rules:
            """
            Process block tokens closing with a new line.
            * rn
            * blockquote
            * codec
            * list
            Notice: Do not processing it when located within a code block, like this:
            if (self.is_in_code() and tag not in {'codec'}):
                continue
            Causing a "jumping" syntax, so better to leave the blocks within the code block
            but re-write their style accordingly.
            """
            if tag not in self.get_nl_closing_tokens():
                continue

            if tag not in self.tokens:
                self.tokens[tag] = {'cnt': 0, 'o': False}

            if self.rehighlight_block:
                if tag in self.tokens:
                    """
                    Opened tokens continue to apply for the following inline parts
                    Logic slightly differ apart with open-close tokens approach.
                    """
                    self.tokens[tag]['o'] = self.user_data.get_param(tag, 'within')

            matches = re.finditer(pattern, text_str)
            match = next(matches, None)

            if match:
                # Get regex result position into the text string
                start = match.start(nth)
                end = match.end(nth)
                length = end - start

                # Collect line tokens only when any of them matched
                if tag not in self.line_tokens[self.line_number]:
                    self.line_tokens[self.line_number][tag] = []
                # The line tokens data will be reset after re-highlighting, no need to check for duplicates
                line_token_data = {'start': start, 'end': end, 'length': length}
                if line_token_data not in self.line_tokens[self.line_number][tag]:
                    self.line_tokens[self.line_number][tag].append(line_token_data)

                if tag == 'rn':
                    self.logger.debug(
                        '{%r}  %s  New line found "%s" at [%d*]'
                        % (self.rehighlight_block, b'\xe2\x86\xb5'.decode('utf-8'), tag, self.line_number_log)
                    )
                    continue

                if ((tag in {'codec', 'list'}
                     # Code block :::: has to start with empty line or at very beginning of the document.
                     and (self.line_number == 0
                          # Check prev line is an empty line
                          or (self.line_number - 1 in self.line_tokens
                              and 'rn' in self.line_tokens[self.line_number - 1])))
                        # Blockquote and list may start without preliminary empty line
                        or (tag not in {'codec'}
                            and not self.tokens[tag]['o'])):
                    # Mind the indents
                    self.tokens[tag]['cnt'] += 1
                    self.tokens[tag]['o'] = True
                    self.logger.debug(
                        '{%r} >>> Open "%s" at [%d*]'
                        % (self.rehighlight_block, tag, self.line_number_log)
                    )
                    """
                    Block tokens cannot be located on the same line, suppose to find them one per line
                    """
                    self.user_data.put(tag=tag, opened=True, within=True, closed=False)
                else:
                    self.logger.debug(
                        '{%r} >~> Open token inside "%s" at [%d*]'
                        % (self.rehighlight_block, tag, self.line_number_log)
                    )
                    """
                    Block tokens cannot be located on the same line, suppose to find them one per line
                    """
                    self.user_data.put(tag=tag, opened=False, within=self.tokens[tag]['o'], closed=False)
            elif self.tokens[tag]['o']:
                if 'rn' in self.line_tokens[self.line_number]:
                    # Close opened token then
                    self.tokens[tag]['o'] = False
                    self.logger.debug(
                        '{%r} <<< Close "%s" at [%d*] %s'
                        % (self.rehighlight_block, tag, self.line_number_log, self.line_tokens[self.line_number])
                    )
                    """
                    Block tokens cannot be located on the same line, suppose to find them one per line
                    """
                    self.user_data.put(tag=tag, opened=False, within=True, closed=True)
                # Some tokens like a 'list' may be located within no empty lines either above or below,
                # thus always an open tag.
                elif tag not in {'list'}:
                    self.logger.debug(
                        '{%r} >=< Inside "%s" at [%d*], is in code %d (lenient), prev state %d'
                        % (self.rehighlight_block, tag, self.line_number_log, self.is_in_code(force_tag='codec'),
                           self.previousBlockState())
                    )
                    """
                    Block tokens cannot be located on the same line, suppose to find them one per line
                    """
                    self.user_data.put(tag=tag, opened=False, within=True, closed=False)
                    if not self.is_any_formatted():
                        if tag == 'blockquote':
                            self.setFormat(0, len(text_str), self.cf(**self.theme['blockquote']))
                            self.set_formatted('blockquote')
                        elif (tag == 'codec'
                              and self.is_in_code(skip_data=True, force_tag='codec')):
                            self.setFormat(0, len(text_str), self.cf(**self.theme['code_content']))
                            self.set_formatted('code')
            else:
                self.user_data.put(tag=tag, opened=False, within=False, closed=False)
                self.logger.debug('{%r} ... No "%s" [%d*]' % (self.rehighlight_block, tag, self.line_number_log))

        try:
            """
            Try to save block data to allow future processing
            """
            self.current_block.setUserData(self.user_data)
        except (TypeError, RuntimeError, ValueError) as e:
            self.logger.error(f'Cannot setup block data "{self.user_data}", error occurred: {e}')

        format_map = {}  # To apply formatting after the whole line processed
        for pattern, nth, tag, group, duple, cf_data, reckon in self.rules:

            if self.is_in_code() and group not in {'code', 'comment', 'coop', 'rn'}:
                """
                Ignore non-code tags if within a code block
                """
                continue

            if reckon is not None and not reckon(self):
                """
                To check some additional conditions for the token, say:
                if not self.is_in_code() and group in {'comment'}: ...
                """
                continue

            if self.rehighlight_block:
                if tag in self.tokens:
                    """
                    Opened tokens continue to apply for the following inline parts
                    Logic slightly differ apart with open-close tokens approach.
                    """
                    self.tokens[tag]['o'] = self.user_data.get_param(tag, 'within')

            """
            Regular Python regular expression operations instead of QRegularExpression, QRegularExpressionMatchIterator
            and QRegularExpressionMatch as sometimes it doesn't work properly.
            * https://docs.python.org/3/library/re.html
            * https://doc.qt.io/qt-6/qregularexpression.html
            """
            matches = re.finditer(pattern, text_str)

            for match in matches:
                # Get regex result position into the text string
                start = match.start(nth)
                end = match.end(nth)
                length = end - start

                # Collect line tokens only when any of them matched
                if tag not in self.line_tokens[self.line_number]:
                    self.line_tokens[self.line_number][tag] = []
                if tag != 'code' or tag not in self.get_nl_closing_tokens():
                    # The line tokens data will be reset after re-highlighting, no need to check for duplicates
                    line_token_data = {'start': start, 'end': end, 'length': length}
                    if line_token_data not in self.line_tokens[self.line_number][tag]:
                        self.line_tokens[self.line_number][tag].append(line_token_data)

                # Check if current tag should be skipped
                if ('codel' in self.line_tokens[self.line_number]
                        and self.line_tokens[self.line_number]['codel']
                        and group not in {'code', 'comment'}
                        and self.pos_within_inline_code(start, end, length)):
                    # skip formatting within inline code block
                    continue

                # Adjust previously processed matches (if rules order is not working)
                if ('codel' in self.line_tokens[self.line_number]
                        and group not in {'code', 'comment'}):
                    self.adjust_pos_within_inline_code()

                # Skip inline code block located within the multi-line code block
                if tag == 'codel' and self.is_in_code():
                    continue

                if tag not in {'code'}:
                    # Set up the tag count
                    if tag in self.tokens:
                        self.tokens[tag]['cnt'] += 1
                    else:
                        self.tokens[tag] = {'cnt': 1, 'o': (True if duple and tag in open_tokens else None)}

                    # If the tag is duple mark it either opened or closed
                    if duple:
                        if tag in open_tokens:
                            self.tokens[tag]['o'] = True
                        elif tag in close_tokens:
                            for _r in self.get_open_close_token_map():
                                if _r['group'] == group and _r['close'] == tag and _r['open'] in self.tokens:
                                    self.tokens[_r['open']]['o'] = False

                # Comment """ block
                if tag == 'cclop' and self.is_in_code():
                    if 'o' not in self.tokens[tag]:
                        self.tokens[tag]['o'] = True
                    elif self.tokens[tag]['o']:
                        self.tokens[tag]['o'] = False
                    else:
                        self.tokens[tag]['o'] = True

                # Table
                if (tag == 'table_h'
                    and (self.line_number - 1 in self.line_tokens
                         # Previous token is a table header
                         and (self.line_number - 1 in self.line_tokens
                              and 'table_d' in self.line_tokens[self.line_number - 1])
                         # The token before the table header either a new line or file's first line
                         and ((self.line_number - 2 in self.line_tokens
                              and 'rn' in self.line_tokens[self.line_number - 2])
                              # line number "1" as table header token earliest line is next to "0"
                              or self.line_number == 1))):
                    self.tokens['table_d']['o'] = True
                    # It will be highlighted after the next re-highlight
                    if self.prev_user_data:
                        self.prev_user_data.put(tag='table_d', opened=True, within=True, closed=False)
                elif 'table_d' in self.tokens and self.tokens['table_d']['o'] and tag != 'table_d':
                    self.tokens['table_d']['o'] = False
                    if self.prev_user_data:
                        self.prev_user_data.put(tag='table_d', opened=False, within=True, closed=True)
                elif tag == 'table_d' and self.tokens['table_d']['o']:
                    self.user_data.put(tag=tag, opened=False, within=True, closed=False)
                if (group == 'table'
                        # Skip if not in table context yet
                        and not (self.user_data.get_param('table_d', 'within')
                                 or ('table_d' in self.tokens and self.tokens['table_d']['o']))):
                    self.logger.debug('Skipping table block')
                    continue
                # Not saving the block's data here;
                # it will be automatically stored during the next block re-highlighting iteration.

                # Header tags correction (emoji)
                if group == 'h':
                    if end == len(text_str):
                        length += 1

                # Prevent passing reference of the dict
                cfc = cf_data.copy()

                # Within the code block
                if self.is_in_code():
                    # Search whether line within code block or not
                    # or True in [True if x['in_code'] else False for x in self.line_tokens[self.line_number][tag]]):
                    if tag == 'code_lang':
                        cfc = self.theme['code_lang']
                    elif tag == 'code_indent':
                        cfc = self.theme['code_indent']
                    # Highlight the wrong indent when detected
                    elif tag == 'wrong_indent':
                        cfc = self.theme['wrong_indent']
                    # Comment (to avoid overriding with code formatting)
                    elif group == 'comment' or self.is_in_code_comment():
                        cfc = self.theme['comment']
                    # elif group == 'coop':
                    #    self.logger.debug('Code operator match', self.line_number_log, match.group(1),
                    #                      start, end, length)
                else:
                    # To avoid highlighting code lang or code text content when not in code
                    # Note: Closing code tag may appear here
                    if (group in {'code', 'comment'}
                            # Closing tag is a valid case
                            and not (tag == 'code'
                                     and self.prev_user_data
                                     and self.prev_user_data.get_param('code', 'within'))
                            # Code line is a valid case of a "not in a code block" tag here
                            and tag not in {'codel'}):
                        # self.logger.debug('Code tag "%s" (group "%s") at [%d] within a non-code context'
                        #                   % (tag, group, self.line_number_log))
                        continue
                    """
                    May causing "jumping" when formatted lines become unformatted,
                    most likely because of compete rules.
                    """
                    if (group == 'list' and not self.is_in_list()
                            # Or list indent is located within a blockquote
                            or tag == 'list_indent' and self.is_in_blockquote()):
                        continue
                    if group == 'blockquote' and not self.is_in_blockquote():
                        continue

                    """
                    Open-close tokens correction.
                    It happens because of matched block started from zero point or so,
                    which is from where the not included element starts, hence the correction.
                    Include optional element within the match brackets to get an actual start position.
                    """
                    if start != 0 and (group in oct_groups or group == 'code'):
                        # Regex's result from 2nd nth
                        if (tag == 'iu'
                                or tag == 'boo'
                                or tag == 'biu'):
                            # If the end of the line than add extra one
                            if end == len(text_str):
                                length += 1
                        elif (tag == 'code_lang' or tag == 'codel' or tag == 'codec'
                              or tag == 'i' or tag == 'i_open' or tag == 'iu_open'
                              or tag == 'b' or tag == 'b_open' or tag == 'boo_open'
                              or tag == 'bi' or tag == 'bi_open' or tag == 'biu_open'
                              or tag == 's' or tag == 's_open'
                              or tag == 'u' or tag == 'u_open'):
                            start = start
                        else:
                            start += 1

                """
                Check is single token inside the open close block of the same group
                """
                opened_group_token = self.get_opened_group_token(group)
                if not duple and opened_group_token is not None:
                    # Have to be there if it is not None
                    self.tokens[opened_group_token]['o'] = False
                    self.tokens[opened_group_token]['cnt'] -= 1

                """
                Non-default encodings may appear, decode them with unknown replacement.
                Also, check `surrogateescape` option.
                E.g. match.captured(nth).encode('utf-8', 'replace').decode('utf-8').
                Also, check match.capturedTexts()
                """
                self.logger.debug(
                    'Tokens: "%s" > %s > %s (s:%d, l:%d, e:%d, n:%d)[%d] prev block state %d'
                    % (tag, self.tokens[tag], match.group(nth).encode('utf-8', 'replace'),
                       start, length, end, nth, self.line_number_log, self.previousBlockState()))

                """
                Inheriting a bg color of the blockquote.
                To keep an element's background use:
                and 'bg' not in cfc
                """
                if (self.is_in_blockquote()
                        and self.is_blockquote_inner(tag)
                        and 'bg' in self.theme['blockquote']):
                    cfc['bg'] = self.theme['blockquote']['bg_inner']

                # QTextCharFormat
                tc_fmt = self.cf(**cfc)
                if tc_fmt is not None:
                    if tag not in format_map:
                        format_map[tag] = []
                    # To apply it later when all line tokens are collected and adjusted
                    format_map[tag].append({'group': group, 'start': start, 'length': length, 'fmt': tc_fmt})

        for tag, fmt_data in format_map.items():
            if tag not in self.line_tokens[self.line_number]:
                continue
            for fd in fmt_data:
                self.setFormat(fd['start'], fd['length'], fd['fmt'])
                self.set_formatted(fd['group'])

        # TODO uncomment later when such block processing will be updated
        # # If within a comment and no other formatting except the comment
        # if self.is_in_code() and self.is_in_code_comment():
        #    self.setFormat(0, len(text_str), self.cf(**self.theme['comment']))
        #    self.set_formatted('comment')

        for token_data in self.get_open_close_token_map():
            """
            Process the lines located between the tags
            """
            # Not formatted elements within the multiline block
            if (not self.is_any_formatted()
                    and not self.is_in_code()
                    # TODO Check and refactor as one group replaces all the others
                    # and not self.is_group_formatted(token_data['group'])
                    and token_data['open'] in self.tokens
                    and self.tokens[token_data['open']]['o'] is True):
                self.setFormat(0, len(text_str), self.cf(**self.theme[token_data['theme']]))
                self.set_formatted(token_data['group'])

    def check_and_set_in_code_state(self):
        if self.is_in_code(True):
            self.logger.debug('In code context')
            if self.currentBlockState() != 1:
                self.setCurrentBlockState(1)
                return True
        else:
            if self.currentBlockState() == 1:
                self.setCurrentBlockState(0)
                return True
        return False

    def pos_within_inline_code(self, start, end, length):
        """
        Check if args position intersects with inline code block.
        """
        for _data in self.line_tokens[self.line_number]['codel']:
            if (_data['start'] < start < _data['end']
                    or _data['start'] < (start + length) < _data['end']
                    or _data['start'] < end < _data['end']):
                # skip formatting within inline code block
                return True

    def adjust_pos_within_inline_code(self) -> bool:
        processed_res = False
        data_to_del = {}
        for _tag, _tag_data in self.line_tokens[self.line_number].items():
            data_to_del[_tag] = []
            for _data in _tag_data:
                if self.pos_within_inline_code(_data['start'], _data['end'], _data['length']):
                    data_to_del[_tag].append(_data)
                    processed_res = True
        for _tag, _tag_data in data_to_del.items():
            for _data in _tag_data:
                self.line_tokens[self.line_number][_tag] = \
                    [d for d in self.line_tokens[self.line_number][_tag] if d != _data]
                if not self.line_tokens[self.line_number][_tag]:
                    del self.line_tokens[self.line_number][_tag]
        # self.logger.debug('Adjusted line tokens:', self.line_tokens[self.line_number])
        return processed_res

    # Check if within a code block
    def is_in_code(self, skip_data=False, force_tag=None):
        if (force_tag
                and force_tag in self.tokens
                and self.tokens[force_tag]['o'] is True):
            return True
        if (not force_tag
                and 'code' in self.tokens
                and self.tokens['code']['o'] is True):
            return True
        elif (not force_tag
              and 'codec' in self.tokens
              and self.tokens['codec']['o'] is True):
            return True
        elif (not skip_data
              and self.currentBlockState() == 1):
            return True
        elif (not skip_data
              and self.user_data is not None
              and self.user_data.get_param('code', 'within')):
            return True
        elif (not skip_data
              and self.user_data is not None
              and self.user_data.get_param('codec', 'within')):
            return True
        else:
            return False

    def is_in_code_comment(self):
        """
        Check if within a comment block
        """
        if ((('comment' in self.tokens and self.tokens['comment']['o'] is True)
             # One line comments
             or 'comment' in self.line_tokens[self.line_number])
                and self.is_in_code()):
            return True
        elif ('cclop' in self.tokens
              and self.tokens['cclop']['o'] is True
              and self.is_in_code()):
            return True
        else:
            return False

    def is_in_blockquote(self, skip_data=False):
        """
        Check if within a blockquote block
        """
        if ('blockquote' in self.tokens
                and self.tokens['blockquote']['o'] is True
                and not self.is_in_code()):
            return True
        elif self.user_data is not None and self.user_data.get_param('blockquote', 'within') and not skip_data:
            return True
        else:
            return False

    def is_in_list(self, skip_data=False):
        """
        Check if within a list
        """
        if ('list' in self.tokens
                and self.tokens['list']['o'] is True
                and not self.is_in_code()):
            return True
        elif self.user_data is not None and self.user_data.get_param('list', 'within') and not skip_data:
            return True
        else:
            return False



# Georgian lexemes toolbar.py
lexemes = {
    "actions_label_new_document": "ახალი დოკუმენტი",
    "actions_accessible_name_new_document": "ახალი დოკუმენტი",
    "actions_label_open": "გახსნა",
    "actions_accessible_name_open": "დოკუმენტის გახსნა",
    "actions_label_save": "შენახვა",
    "actions_accessible_name_save": "დოკუმენტის შენახვა",
    "actions_label_save_as": "შენახვა როგორც",
    "actions_accessible_name_save_as": "დოკუმენტის ასლის შენახვა",
    "actions_label_sync": "სინქრონიზაცია",
    "actions_accessible_name_sync": "დოკუმენტის სინქრონიზაცია გარე მეხსიერებით",
    "actions_label_edit": "რედაქტირება",
    "actions_accessible_name_edit": "ფაილის რედაქტირება",
    "actions_label_source": "წყარო",
    "actions_accessible_name_source": "კოდის წყარო",
    "actions_label_encrypt": "დაშიფვრა",
    "actions_accessible_name_encrypt": "ფაილის დაშიფვრა",
    "actions_label_decrypt": "გაშიფვრა",
    "actions_accessible_name_decrypt": "ფაილის გაშიფვრა",
    "actions_label_bold": "მუქი",
    "actions_accessible_name_bold": "ტექსტის მუქი ფორმატი",
    "actions_label_italic": "კურსივი",
    "actions_accessible_name_italic": "ტექსტის კურსივის ფორმატი",
    "actions_label_underline": "ხაზგასმა",
    "actions_accessible_name_underline": "ტექსტის ხაზგასმის ფორმატი",
    "actions_label_strikethrough": "გადახაზვა",
    "actions_accessible_name_strikethrough": "ტექსტის გადახაზული ფორმატი",
    "actions_label_blockquote": "ბლოკციტატა",
    "actions_accessible_name_blockquote": "ბლოკციტატის ტექსტის ფორმატი",
    "actions_label_color_picker": "ფერის შერჩევა",
    "actions_accessible_name_color_picker": "ტექსტის ფერის არჩევა",
    "actions_label_ai_assistant": "AI ასისტენტი",
    "actions_accessible_name_ai_assistant": "AI ასისტენტის გაშვება",
    "actions_label_about": "შესახებ",
    "actions_accessible_name_about": "აპლიკაციის შესახებ",
    "actions_label_exit": "გასვლა",
    "actions_accessible_name_exit": "აპლიკაციის დატოვება",
    "actions_label_settings": "პარამეტრები",
    "actions_accessible_name_settings": "აპლიკაციის პარამეტრები",

    "search_input_label": "ძიება",
    "search_input_placeholder_text": "ტექსტის ძიება დოკუმენტში",
    "search_input_accessible_description": "ძიების ველი",
    "search_buttons_label_clear": "ძიების გასუფთავება",
    "search_buttons_accessible_name_clear": "ძიების ველის გასუფთავება",
    "search_buttons_label_prev": "წინა ძიება",
    "search_buttons_accessible_name_prev": "წინა ძიების შედეგი",
    "search_buttons_label_next": "შემდეგი ძიება",
    "search_buttons_accessible_name_next": "შემდეგი ძიების შედეგი",
    "search_case_sensitive_accessible_name": "რეგისტრის მნიშვნელობა",
    "search_case_sensitive_tooltip": "რეგისტრის მნიშვნელობით ძიება",
}



# Japanese lexemes settings_dialog.py
lexemes = {
    # 設定
    "window_title": "設定",

    "button_close": "閉じる",

    "tab_general": "一般",
    "tab_editor_config": "エディター",
    "tab_viewer_config": "ビューア",
    "tab_ai_config": "AI 設定",

    "general_app_config_label": "アプリ設定",
    "general_app_language_label": "言語",
    "general_app_language_combo_placeholder_text": "言語を選択",
    "general_app_language_combo_accessible_description": "アプリのインターフェース言語",
    "general_app_theme_label": "テーマ",
    "general_app_theme_combo_placeholder_text": "テーマを選択",
    "general_app_theme_combo_accessible_description": "アプリのインターフェーステーマ",
    "general_app_main_menu_label": "メインメニュー",
    "general_app_main_menu_checkbox": "メインメニューを表示",
    "general_app_main_menu_checkbox_accessible_description": "アプリのメインドロップダウンメニューを表示",
    "general_app_font_size_label": "フォントサイズ：{size}pt",
    "general_app_font_size_slider_accessible_description": "アプリの全体的なフォントサイズを調整する",

    "general_statusbar_label": "ステータスバー",
    "general_statusbar_show_global_cursor_position_checkbox": "グローバルカーソル位置を表示",
    "general_statusbar_show_global_cursor_position_checkbox_accessible_description":
        "ステータスバーにグローバルカーソル位置を表示する",

    "editor_config_label": "エディタ設定",
    "editor_config_show_line_numbers_checkbox": "行番号を表示",
    "editor_config_show_line_numbers_checkbox_accessible_description": "エディタ内で行番号を表示する",

    "viewer_config_label": "ビューワ設定",
    "viewer_config_process_emojis_checkbox": "テキスト絵文字をグラフィックに変換",
    "viewer_config_process_emojis_checkbox_accessible_description": "テキスト絵文字をグラフィカルな表現に変換する",
    "viewer_config_highlight_todos_checkbox": "TODOをハイライト",
    "viewer_config_highlight_todos_checkbox_accessible_description": "テキスト内のTODOタグを強調表示する",
    "viewer_config_open_link_confirmation_checkbox": "リンク開く前に確認が必要",
    "viewer_config_open_link_confirmation_checkbox_accessible_description": "リンクを開く前に確認を求める",
    "viewer_config_save_resources_checkbox": "外部画像をディスクに自動保存",
    "viewer_config_save_resources_checkbox_accessible_description":
        "オフラインアクセスのために外部画像のコピーをディスクに自動保存します。",

    "ai_config_inference_module_label": "推論モジュール",
    "ai_config_inference_module_names_combo_label": "アクティブ推論モジュール",
    "ai_config_inference_module_names_combo_placeholder_text": "モジュールを選択",
    "ai_config_inference_module_names_combo_accessible_description":
        "AIアシスタントと連携して動作する利用可能なAI推論モジュールから選択してください。\n"
        "オプションには、リアルタイム処理を備えたローカルの大規模言語モデル（LLM）またはAPIベースの機能が含まれます。",

    "ai_config_base_label": "基本パラメーター",
    "ai_config_multi_turn_dialogue_checkbox": "会話メモリを伴うマルチターンダイアログ",
    "ai_config_multi_turn_dialogue_checkbox_accessible_description":
        "以前のプロンプトを保持する会話メモリ付きのマルチターンダイアログを有効にします。\n"
        "オフの場合、新しいメッセージとシステムプロンプトのみが応答に影響します。",
    "ai_config_convert_to_md_checkbox": "結果をMarkdownに変換",
    "ai_config_convert_to_md_checkbox_accessible_description":
        "出力メッセージをMarkdown形式に変換します。",
}



"""
Notolog Editor
An open-source Markdown editor built with Python.

File Details:
- Purpose: Contains unit and integration tests for the related functionality.

Repository: https://github.com/notolog/notolog-editor
Website: https://notolog.app
PyPI: https://pypi.org/project/notolog

Author: Vadim Bakhrenkov
Copyright: 2024-2025 Vadim Bakhrenkov
License: MIT License

For detailed instructions and project information, please see the repository's README.md.
"""

from PySide6.QtCore import QEventLoop
from PySide6.QtWidgets import QApplication, QStyleFactory

from notolog.app import main
from notolog.app_config import AppConfig
from notolog.notolog_editor import NotologEditor

from types import SimpleNamespace

from unittest.mock import AsyncMock

import sys
import pytest
import logging
import asyncio


class TestApp:

    @pytest.fixture(scope="function")
    def test_obj_app_config(self, mocker):
        # Mock AppConfig's get_logger_level method to suppress logging during tests.
        mocker.patch.object(AppConfig, 'get_logger_level', return_value=logging.NOTSET)

        _app_config = AppConfig()
        _app_config.set_test_mode(True)

        yield _app_config

    @pytest.fixture(scope="function")
    def test_exp_params_fixture(self, request):
        # Retrieve parameter values from the test request.
        param_values = request.param

        yield param_values

    @pytest.mark.parametrize(
        "test_exp_params_fixture",
        [
            ('-smth1', '', 2),
            ('-v', f'Notolog {AppConfig().get_app_version()}', 0),
            ('--version', f'Notolog {AppConfig().get_app_version()}', 0),
        ],
        indirect=True
    )
    def test_main_args(self, test_obj_app_config: AppConfig, test_exp_params_fixture, monkeypatch, capsys):
        arg, exp_result, exp_exit_code = test_exp_params_fixture

        # Simulate command-line arguments
        monkeypatch.setattr(sys, "argv", ['app.py', arg])

        # Call the main function
        with pytest.raises(SystemExit) as excinfo:
            main()

        # Validate the output and the exit code
        captured = capsys.readouterr()
        assert captured.out.strip() == exp_result
        assert excinfo.value.code == exp_exit_code

    @pytest.mark.asyncio
    async def test_app(self, test_obj_app_config: AppConfig, mocker, monkeypatch):
        # Simulate command-line arguments
        monkeypatch.setattr(sys, "argv", [])

        event_loop = mocker.patch.object(QEventLoop, '__init__', return_value=None)
        mocker.patch.object(asyncio, 'set_event_loop', return_value=None)
        # AsyncMock mock class is designed to simulate asynchronous calls and can be awaited.
        mocker.patch.object(asyncio, 'Event', return_value=SimpleNamespace(**{'set': lambda: None, 'wait': AsyncMock()}))
        mocker.patch.object(event_loop, 'run_until_complete', return_value=None)

        test_notolog_editor_show = mocker.patch.object(NotologEditor, 'show', wraps=lambda: None)
        # Prevent resource processing, including 'process_document_images'
        mocker.patch.object(NotologEditor, 'load_content_html', return_value=None)
        test_logging_basic_config = mocker.patch.object(logging, 'basicConfig', return_value=None)
        test_set_organisation_name = mocker.patch.object(QApplication, 'setOrganizationName')
        test_set_organisation_domain = mocker.patch.object(QApplication, 'setOrganizationDomain')
        test_set_application_name = mocker.patch.object(QApplication, 'setApplicationName')
        test_set_application_version = mocker.patch.object(QApplication, 'setApplicationVersion')
        test_set_desktop_settings_aware = mocker.patch.object(QApplication, 'setDesktopSettingsAware')
        test_qstylefactory_create = mocker.patch.object(QStyleFactory, 'create', wraps=QStyleFactory.create)
        test_set_style = mocker.patch.object(QApplication, 'setStyle')

        # Call the main function
        main()

        test_logging_basic_config.assert_called_once()

        test_set_organisation_name.assert_called_once()
        assert str(test_set_organisation_name.call_args) == "call('%s')" % test_obj_app_config.get_settings_org_name()

        test_set_organisation_domain.assert_called_once()
        assert (str(test_set_organisation_domain.call_args) == "call('%s')" % test_obj_app_config.get_settings_org_domain())

        test_set_application_name.assert_called_once()
        assert (str(test_set_application_name.call_args) == "call('%s')" % test_obj_app_config.get_settings_app_name())

        test_set_application_version.assert_called_once()
        assert (str(test_set_application_version.call_args) == "call('%s')" % test_obj_app_config.get_app_version())

        test_set_desktop_settings_aware.assert_called_once()
        assert (str(test_set_desktop_settings_aware.call_args) == "call(%s)" % 'False')

        test_qstylefactory_create.assert_called_once()
        assert (str(test_qstylefactory_create.call_args) == "call('%s')" % 'Fusion')

        test_set_style.assert_called_once()

        test_notolog_editor_show.assert_called()

    @pytest.mark.asyncio
    async def test_async(self):
        assert await asyncio.sleep(0, result=True)



# Spanish lexemes ai_assistant.py
lexemes = {
    # Asistente IA
    "dialog_title": "Solicitud a IA",
    "dialog_prompt_input_placeholder_text": "Solicitud",
    "dialog_prompt_input_accessible_description": "Ingrese la solicitud para dar un contexto al asistente IA",
    "dialog_response_output_placeholder_text": "Datos de respuesta",
    "dialog_response_output_accessible_description": "Los datos de respuesta aparecerán en este campo",
    "dialog_response_output_notice_empty_text":
        "Proporcione una entrada de texto para obtener una respuesta de aproximadamente...",

    "dialog_usage_module_label": "MÓDULO",
    "dialog_usage_model_label": "MODELO",
    "dialog_usage_tokens_label": "TOKENS",
    "dialog_usage_tokens_prompt": "prompt: {tokens}",
    "dialog_usage_tokens_answer": "respuesta: {tokens}",
    "dialog_usage_tokens_total": "total: {tokens}",

    "dialog_button_send_request": "Enviar solicitud",
    "dialog_button_stop_request": "Detener",
    "dialog_button_save_history": "Guardar historial del diálogo",

    "dialog_message_copy_tooltip": "Copiar texto",
    "dialog_message_copied_tooltip": "Copiado",

    "dialog_error_loading_model": "Error al cargar el modelo personalizado",
}



# Russian lexemes settings_dialog.py
lexemes = {
    "tab_module_llama_cpp_config": "Модуль llama.cpp",

    "module_llama_cpp_config_label": "Модуль llama.cpp",
    "module_llama_cpp_config_path_label": "Расположение модели",
    "module_llama_cpp_config_path_input_placeholder_text": "Выберите или введите путь к модели",
    "module_llama_cpp_config_path_input_accessible_description":
        "Поле ввода с селектором для указания пути локальной модели. Поддерживает модели в формате GGUF,\n"
        "бинарный формат файла, оптимизированный для хранения моделей, используемых с GGML и основанными на GGML "
        "исполнителями.",
    "module_llama_cpp_config_path_input_filter_text": "Файлы GGUF",

    "module_llama_cpp_config_context_window_label": "Размер окна контекста",
    "module_llama_cpp_config_context_window_input_accessible_description":
        "Устанавливает количество токенов, которые модель учитывает при генерации ответов.\n"
        "Контролирует, сколько предыдущего контекста используется.",

    "module_llama_cpp_chat_formats_label": "Форматы чата",
    "module_llama_cpp_chat_formats_combo_placeholder_text": "Выберите формат чата",
    "module_llama_cpp_chat_formats_combo_accessible_description":
        "Выпадающее меню для выбора формата, используемого для разговоров модели.",

    "module_llama_cpp_config_system_prompt_label": "Системный промпт",
    "module_llama_cpp_config_system_prompt_edit_placeholder_text": "Введите текст системного промпта",
    "module_llama_cpp_config_system_prompt_edit_accessible_description":
        "Текстовое поле для ввода системного промпта, который направляет ответы модели.",

    "module_llama_cpp_config_response_temperature_label": "Температура ответа: {temperature}",
    "module_llama_cpp_config_response_temperature_input_accessible_description":
        "Регулирует случайность ответов модели. Высокие значения создают более разнообразные результаты,\n"
        "в то время как низкие значения приводят к более предсказуемым ответам.",

    "module_llama_cpp_config_response_max_tokens_label": "Максимальное количество токенов на ответ",
    "module_llama_cpp_config_response_max_tokens_input_accessible_description":
        "Ограничивает количество токенов в ответах модели до фактического предела окна контекста.\n"
        "Нулевое значение предполагает емкость окна контекста.",

    "module_llama_cpp_config_prompt_history_size_label": "Размер истории",
    "module_llama_cpp_config_prompt_history_size_input_accessible_description":
        "Контролирует количество записей в истории сохраняемых системой.\n"
        "Нулевое значение позволяет неограниченное количество записей."
}



# Italian lexemes settings_dialog.py
lexemes = {
    "tab_ondevice_llm_config": "LLM su Dispositivo",

    "module_ondevice_llm_config_label": "Modello LLM su Dispositivo",
    "module_ondevice_llm_config_path_label": "Posizione del Modello ONNX",
    "module_ondevice_llm_config_path_input_placeholder_text": "Percorso della directory del modello",
    "module_ondevice_llm_config_path_input_accessible_description":
        "Un campo di input con un selettore per specificare il percorso della directory del modello dove si trovano\n"
        "i file ONNX. I modelli supportati sono nel formato ONNX, che sta per Open Neural Network Exchange,\n"
        "uno standard aperto per i formati dei modelli di apprendimento automatico.",

    "module_ondevice_llm_config_response_temperature_label": "Temperatura: {temperature}",
    "module_ondevice_llm_config_response_temperature_input_accessible_description":
        "Regola la casualità delle risposte del modello. Valori più alti producono output più variati,\n"
        "mentre valori più bassi rendono le risposte più prevedibili.",

    "module_ondevice_llm_config_response_max_tokens_label": "Massimo di Token per Risposta",
    "module_ondevice_llm_config_response_max_tokens_input_accessible_description":
        "Imposta il numero massimo di token da ricevere in una risposta, come parole e punteggiatura,\n"
        "controllando la lunghezza dell'output.",

    "module_ondevice_llm_config_prompt_history_size_label": "Dimensione della Cronologia dei Prompt",
    "module_ondevice_llm_config_prompt_history_size_input_accessible_description":
        "Controlla il numero di voci nella cronologia dei prompt che il sistema conserva per riferimento.\n"
        "Un valore zero consente un numero illimitato di voci."
}



# Russian lexemes toolbar.py
lexemes = {
    "actions_label_new_document": "Новый документ",
    "actions_accessible_name_new_document": "Новый документ",
    "actions_label_open": "Открыть",
    "actions_accessible_name_open": "Открыть документ",
    "actions_label_save": "Сохранить",
    "actions_accessible_name_save": "Сохранить документ",
    "actions_label_save_as": "Сохранить как",
    "actions_accessible_name_save_as": "Сохранить копию документа",
    "actions_label_sync": "Синхронизировать",
    "actions_accessible_name_sync": "Синхронизировать документ с внешним хранилищем",
    "actions_label_edit": "Редактировать",
    "actions_accessible_name_edit": "Редактировать файл",
    "actions_label_source": "Источник",
    "actions_accessible_name_source": "Исходный код",
    "actions_label_encrypt": "Зашифровать",
    "actions_accessible_name_encrypt": "Зашифровать файл",
    "actions_label_decrypt": "Расшифровать",
    "actions_accessible_name_decrypt": "Расшифровать файл",
    "actions_label_bold": "Жирный",
    "actions_accessible_name_bold": "Жирный формат текста",
    "actions_label_italic": "Курсив",
    "actions_accessible_name_italic": "Курсивный формат текста",
    "actions_label_underline": "Подчеркнутый",
    "actions_accessible_name_underline": "Подчеркнутый формат текста",
    "actions_label_strikethrough": "Зачеркнутый",
    "actions_accessible_name_strikethrough": "Зачеркнутый формат текста",
    "actions_label_blockquote": "Цитата",
    "actions_accessible_name_blockquote": "Формат текста в виде цитаты",
    "actions_label_color_picker": "Выбор цвета",
    "actions_accessible_name_color_picker": "Выберите цвет текста",
    "actions_label_ai_assistant": "ИИ помощник",
    "actions_accessible_name_ai_assistant": "Запустить ИИ помощника",
    "actions_label_about": "О программе",
    "actions_accessible_name_about": "О приложении",
    "actions_label_exit": "Выход",
    "actions_accessible_name_exit": "Выйти из приложения",
    "actions_label_settings": "Настройки",
    "actions_accessible_name_settings": "Настройки приложения",

    "search_input_label": "Поиск",
    "search_input_placeholder_text": "Поиск текста в документе",
    "search_input_accessible_description": "Поле ввода поиска",
    "search_buttons_label_clear": "Очистить поиск",
    "search_buttons_accessible_name_clear": "Очистить поле поиска",
    "search_buttons_label_prev": "Предыдущий поиск",
    "search_buttons_accessible_name_prev": "Предыдущий результат поиска",
    "search_buttons_label_next": "Следующий поиск",
    "search_buttons_accessible_name_next": "Следующий результат поиска",
    "search_case_sensitive_accessible_name": "Учитывать регистр",
    "search_case_sensitive_tooltip": "Поиск с учетом регистра",
}



# Swedish lexemes main_menu.py
lexemes = {
    "group_file_label": "Fil",
    "group_edit_label": "Redigera",
    "group_tools_label": "Verktyg",
    "group_help_label": "Hjälp",

    "actions_file_label_new_document": "Nytt dokument",
    "actions_file_accessible_name_new_document": "Nytt dokument",
    "actions_file_label_open": "Öppna",
    "actions_file_accessible_name_open": "Öppna dokument",
    "actions_file_label_save": "Spara",
    "actions_file_accessible_name_save": "Spara dokument",
    "actions_file_label_save_as": "Spara som",
    "actions_file_accessible_name_save_as": "Spara dokumentkopior",
    "actions_file_label_settings": "Inställningar",
    "actions_file_accessible_name_settings": "Appinställningar",
    "actions_file_label_reset_settings": "Återställ inställningar",
    "actions_file_accessible_name_reset_settings": "Återställ inställningar",
    "actions_file_label_exit": "Avsluta",
    "actions_file_accessible_name_exit": "Avsluta programmet",

    "actions_edit_edit_mode": "Redigera",
    "actions_edit_accessible_name_edit_mode": "Redigera fil",
    "actions_edit_source_mode": "Källkod",
    "actions_edit_accessible_name_source_mode": "Källkod",
    "actions_edit_bold": "Fet",
    "actions_edit_accessible_name_bold": "Fet textformat",
    "actions_edit_italic": "Kursiv",
    "actions_edit_accessible_name_italic": "Kursiv textformat",
    "actions_edit_underline": "Understrykning",
    "actions_edit_accessible_name_underline": "Understruken textformat",
    "actions_edit_strikethrough": "Genomstrykning",
    "actions_edit_accessible_name_strikethrough": "Genomstruken textformat",
    "actions_edit_blockquote": "Blockcitat",
    "actions_edit_accessible_name_blockquote": "Blockcitat textformat",

    "actions_tools_ai_assistant": "AI-assistent",
    "actions_tools_accessible_name_ai_assistant": "Starta AI-assistent",
    "actions_tools_color_picker": "Färgväljare",
    "actions_tools_accessible_name_color_picker": "Välj textfärg",

    "actions_help_label_md_syntax": "Markdown-syntax",
    "actions_help_accessible_name_md_syntax": "Markdown-syntax",
    "actions_help_label_check_for_updates": "Sök efter uppdateringar",
    "actions_help_accessible_name_check_for_updates": "Sök efter uppdateringar",
    "actions_help_label_bug_report": "Felrapport",
    "actions_help_accessible_name_bug_report": "Skicka felrapport",
    "actions_help_label_about": "Om",
    "actions_help_accessible_name_about": "Om"
}



# Russian lexemes common.py
lexemes = {
    "module_llama_cpp_name": "Модуль llama.cpp",

    "module_llama_cpp_model_exception": "Ошибка: модель не найдена: {error_msg}",
    "module_llama_cpp_task_exception": "Невозможно инициировать генератор модели: {error_msg}",
}



# Spanish lexemes common.py
lexemes = {
    "app_title": "Editor Notolog",
    "app_title_with_sub": "{app_title} - {sub_title}",

    "tree_filter_accessible_desc": "Campo de filtro de archivos",

    "menu_action_copy_file_path": "Copiar ruta del archivo",
    "menu_action_rename": "Renombrar",
    "menu_action_delete": "Eliminar",
    "menu_action_delete_completely": "Eliminar completamente",
    "menu_action_restore": "Restaurar",
    "menu_action_create_new_dir": "Crear un nuevo directorio",

    "dialog_file_rename_title": "Renombrar archivo",
    "dialog_file_rename_field_label": "Ingrese el nuevo nombre del archivo",
    "dialog_file_rename_button_ok": "Renombrar",
    "dialog_file_rename_warning_exists": "Ya existe un archivo con el mismo nombre",

    "dialog_file_delete_title": "Eliminar archivo",
    "dialog_file_delete_text": "¿Eliminar archivo \"{file_name}\"?",
    "dialog_file_delete_completely_title": "Eliminar archivo completamente",
    "dialog_file_delete_completely_text": "¿Eliminar completamente el archivo \"{file_name}\"?",
    "dialog_file_delete_error": "No se puede eliminar el archivo, ocurrió un error",
    "dialog_file_delete_error_not_found": "Archivo no encontrado",

    "dialog_file_restore_title": "Restaurar archivo",
    "dialog_file_restore_text": "¿Restaurar el archivo \"{file_name}\"?",
    "dialog_file_restore_error": "No se puede restaurar el archivo, ocurrió un error",
    "dialog_file_restore_warning_exists": "Ya existe un archivo con el nombre {file_name}",

    "dialog_create_new_dir_title": "Crear un nuevo directorio",
    "dialog_create_new_dir_label": "Nombre del nuevo directorio",
    "dialog_create_new_dir_input_placeholder_text": "Ingrese el nombre del directorio",
    "dialog_create_new_dir_button_ok": "Crear",
    "dialog_create_new_dir_button_cancel": "Cancelar",
    "dialog_create_new_dir_warning_empty_title": "Error en el nombre del nuevo directorio",
    "dialog_create_new_dir_warning_empty_text": "El nombre del directorio no puede estar vacío",
    "dialog_create_new_dir_warning_too_long_title": "Error en el nombre del nuevo directorio",
    "dialog_create_new_dir_warning_too_long_text": "El nombre del directorio es demasiado largo; ¡máximo "
                                                   "{symbols} caracteres permitidos!",
    "dialog_create_new_dir_error_existed": "El directorio ya existe",
    "dialog_create_new_dir_error": "No se puede crear el directorio. Asegúrese de que el directorio de destino "
                                   "{base_dir} sea escribible",

    "dialog_message_box_title": "Mensaje",
    "dialog_message_box_button_ok": "Cerrar",

    "action_new_file_first_line_template_text": "Nuevo documento",
    "action_open_file_dialog_caption": "Abrir archivo",
    "action_save_as_file_dialog_caption": "Guardar archivo",

    "dialog_save_empty_file_title": "Guardar archivo vacío",
    "dialog_save_empty_file_text": "¿Permitir guardar el archivo con contenido vacío?",

    "dialog_encrypt_file_title": "Encriptar archivo",
    "dialog_encrypt_file_text": "¿Encriptar archivo \"{file_name}\"?",
    "encrypt_file_warning_file_is_already_encrypted": "¡El archivo ya está encriptado!",
    "dialog_encrypt_file_rewrite_existing_title": "Reescribir archivo existente",
    "dialog_encrypt_file_rewrite_existing_text": "¿Reescribir archivo existente \"{file_path}\"?",

    "dialog_decrypt_file_title": "Desencriptar archivo",
    "dialog_decrypt_file_text": "¿Desencriptar archivo \"{file_name}\"?",
    "decrypt_file_warning_file_is_not_encrypted": "¡El archivo no está encriptado!",
    "dialog_decrypt_file_rewrite_existing_title": "Reescribir archivo existente",
    "dialog_decrypt_file_rewrite_existing_text": "¿Reescribir archivo existente \"{file_path}\"?",

    "dialog_encrypt_new_password_title": "Nueva Contraseña",
    "dialog_encrypt_new_password_label": "Contraseña:",
    "dialog_encrypt_new_password_input_placeholder_text": "Introduzca la nueva contraseña",
    "dialog_encrypt_new_password_hint_label": "Pista:",
    "dialog_encrypt_new_password_hint_label_description":
        "¡La pista no está cifrada y puede ser leída desde el archivo! "
        "\nEvite pistas obvias que se puedan adivinar fácilmente, como fechas de nacimiento. "
        "\nIntente utilizar una referencia que no se asocie fácilmente con usted.",
    "dialog_encrypt_new_password_hint_input_placeholder_text": "Introduzca una pista (opcional)",
    "dialog_encrypt_new_password_button_ok": "Aceptar",
    "dialog_encrypt_new_password_button_cancel": "Cancelar",
    "dialog_encrypt_new_password_warning_empty_title": "Advertencia",
    "dialog_encrypt_new_password_warning_empty_text": "¡El campo de contraseña no puede estar vacío!",
    "dialog_encrypt_new_password_warning_too_long_title": "Advertencia",
    "dialog_encrypt_new_password_warning_too_long_text":
        "El campo de la pista es demasiado largo, ¡máximo de {symbols} caracteres!",

    "dialog_encrypt_password_title": "Introducir Contraseña",
    "dialog_encrypt_password_label": "Contraseña:",
    "dialog_encrypt_password_input_placeholder_text": "Introducir contraseña",
    "dialog_encrypt_password_hint_label": "Pista:",
    "dialog_encrypt_password_button_ok": "Aceptar",
    "dialog_encrypt_password_button_cancel": "Cancelar",

    "dialog_encrypt_password_reset_title": "Restablecer Contraseña de Cifrado",
    "dialog_encrypt_password_reset_text": "¿Está seguro de querer restablecer la contraseña de cifrado actual?",
    "dialog_encrypt_password_reset_button_cancel": "Cancelar",
    "dialog_encrypt_password_reset_button_yes": "Sí",

    "dialog_open_link_title": "Enlace",
    "dialog_open_link_text": "¿Abrir enlace \"{url}\" en un navegador?",

    "dialog_reset_settings_title": "¿Restablecer configuraciones?",
    "dialog_reset_settings_text":
        "Todos los datos almacenados en las configuraciones serán borrados, y la aplicación se reiniciará para aplicar "
        "los cambios.",

    "dialog_exit_unsaved_title": "Confirmar Salida",
    "dialog_exit_unsaved_text": "El archivo abierto '{file_name}' no puede ser guardado. ¿Continuar con la salida?",

    "message_app_config_file_access": "Permiso denegado al acceder al archivo de configuración de la aplicación en "
                                      "{file_path}. Establezca los permisos correctos para asegurar un funcionamiento "
                                      "adecuado.",

    "field_dir_path_dialog_caption": "Seleccionar Directorio",
    "field_file_path_dialog_caption": "Seleccionar Archivo",

    "load_file_encryption_password_mismatch": "¡La contraseña de encriptación no coincide!",
    "load_file_encryption_password_incorrect": "¡Contraseña de encriptación incorrecta!",
    "load_file_none_content_error": "No se puede cargar el archivo.",

    "action_new_file_error_occurred": "No se puede crear el archivo; ocurrió un error."
                                      "\nVerifique los permisos del sistema de archivos.",
    "save_active_file_error_occurred": "No se puede guardar el archivo; ocurrió un error.",

    "expandable_block_default_title": "Más información...",
    "expandable_block_open_close_tags_mismatch_warning":
        "Incompatibilidad de etiquetas de apertura/cierre del bloque <details>",

    "dialog_color_picker_color_copied_to_the_clipboard": "El texto formateado ha sido copiado al portapapeles",

    "popup_about_title": "Información de la Aplicación",
    "popup_about_app_name_description": "Editor Markdown de Python",

    "popup_about_version": "Versión",
    "popup_about_license": "Licencia",
    "popup_about_website": "Sitio Web",
    "popup_about_repository": "GitHub",
    "popup_about_pypi": "PyPi",
    "popup_about_date": "Fecha",

    "update_helper_new_version_is_available": "Una nueva versión {latest_version} de la aplicación está disponible",
    "update_helper_latest_version_installed": "La última versión de la aplicación está instalada",

    "network_connection_error_empty": "No se puede obtener información de la respuesta",
    "network_connection_error_connection_or_dns":
        "Host no encontrado. Puede haber un problema con la conexión a internet o DNS.",
    "network_connection_error_connection_refused":
        "Conexión rechazada. El servidor podría estar caído o puede haber problemas de red.",
    "network_connection_error_connection_timed_out": "Tiempo de conexión agotado. Puede haber problemas de red.",
    "network_connection_error_connection_404_error":
        "Error de conexión 404. La página o recurso solicitado no se encuentra.",
    "network_connection_error_generic_with_status_code": "La solicitud falló con el código de estado: {status_code}",
}



# Spanish lexemes common.py
lexemes = {
    # Common
    "module_ondevice_llm_name": "LLM en dispositivo",

    "module_ondevice_llm_model_exception": "Error de modelo no encontrado: {error_msg}",
    "module_ondevice_llm_task_exception": "No se puede iniciar el generador de modelos: {error_msg}",
}



# Russian lexemes settings_dialog.py
lexemes = {
    "tab_openai_api_config": "API OpenAI",

    "module_openai_api_label": "API OpenAI",
    "module_openai_api_url_label": "URL API",
    "module_openai_api_url_input_placeholder_text": "URL API",
    "module_openai_api_url_input_accessible_description":
        "URL API OpenAI — это адрес конечной точки API, который может изменяться в зависимости от сервиса и версии.\n"
        "AI Assistant использует ту, что предназначена для функций чатов или завершения текста.\n"
        "Обратитесь к официальной документации OpenAI API, чтобы получить текущий URL.",
    "module_openai_api_key_label": "Ключ API",
    "module_openai_api_key_input_placeholder_text": "Ключ API",
    "module_openai_api_key_input_accessible_description":
        "Ключ API OpenAI — это секретный токен, используемый для аутентификации запросов к конечной точке API.",
    "module_openai_api_supported_models_label": "Поддерживаемые модели",
    "module_openai_api_model_names_combo_placeholder_text": "Выбрать модель",
    "module_openai_api_model_names_combo_accessible_description":
        "Выберите из поддерживаемых моделей для чат-бесед.",

    "module_openai_api_base_system_prompt_label": "Системный промпт",
    "module_openai_api_base_system_prompt_edit_placeholder_text":
        "Базовый системный промпт, предшествующий каждому запросу",
    "module_openai_api_base_system_prompt_edit_accessible_description":
        "Базовый системный промпт, который предшествует каждому запросу.\n"
        "Обычно это простой текст с инструкциями или характеристиками роли.",

    "module_openai_api_base_response_temperature_label": "Температура: {temperature}",
    "module_openai_api_base_response_temperature_input_accessible_description":
        "Регулирует случайность ответов модели. Высокие значения создают более разнообразные результаты, "
        "в то время как низкие значения делают ответы более предсказуемыми.",

    "module_openai_api_base_response_max_tokens_label": "Максимальное количество токенов на ответ",
    "module_openai_api_base_response_max_tokens_input_accessible_description":
        "Максимальное количество токенов, которые можно получить в ответ, таких как слова и пунктуация, "
        "контролируя длину вывода.",

    "module_openai_api_config_prompt_history_size_label": "Размер истории промптов",
    "module_openai_api_config_prompt_history_size_input_accessible_description":
        "Контролирует количество записей в истории промптов, которые система сохраняет для справки.\n"
        "Значение ноль позволяет неограниченное количество записей."
}
